@@ -53,107 +53,111 @@ record_formatters = [
 
 snapshot_formatters = [
         Formatter('<total bytes>',
-            lambda snapshot, record, raw, fmt, rbargs:
-            _damo_fmt_str.format_sz(snapshot.total_bytes, raw),
+            lambda snapshot, record, fmt:
+            _damo_fmt_str.format_sz(snapshot.total_bytes, fmt.raw_number),
             'total bytes of regions in the snapshot'),
         Formatter('<duration>',
-            lambda snapshot, record, raw, fmt, rbargs:
+            lambda snapshot, record, fmt:
                   _damo_fmt_str.format_time_ns(
-                      snapshot.end_time - snapshot.start_time, raw),
+                      snapshot.end_time - snapshot.start_time, fmt.raw_number),
                   'access monitoring duration for the snapshot'),
         Formatter('<start time>',
-            lambda snapshot, record, raw, fmt, rbargs:
+            lambda snapshot, record, fmt:
                   _damo_fmt_str.format_time_ns(
                       snapshot.start_time -
-                      record.snapshots[0].start_time, raw),
+                      record.snapshots[0].start_time, fmt.raw_number),
             'access monitoring start time for the snapshot, relative to the record start time'),
         Formatter('<end time>',
-            lambda snapshot, record, raw, fmt, rbargs:
+            lambda snapshot, record, fmt:
                   _damo_fmt_str.format_time_ns(
-                      snapshot.end_time - record.snapshots[0].start_time, raw),
+                      snapshot.end_time - record.snapshots[0].start_time,
+                      fmt.raw_number),
             'access monitoring end time for the snapshot, relative to the record end time'),
         Formatter('<abs start time>',
-            lambda snapshot, record, raw, fmt, rbargs:
-            _damo_fmt_str.format_time_ns(snapshot.start_time, raw),
+            lambda snapshot, record, fmt:
+            _damo_fmt_str.format_time_ns(snapshot.start_time, fmt.raw_number),
             'absolute access monitoring start time for the snapshot'),
         Formatter('<abs end time>',
-            lambda snapshot, record, raw, fmt, rbargs:
-            _damo_fmt_str.format_time_ns(snapshot.end_time, raw),
+            lambda snapshot, record, fmt:
+            _damo_fmt_str.format_time_ns(snapshot.end_time, fmt.raw_number),
             'absolute access monitoring end time for the snapshot'),
         Formatter('<number of regions>',
-            lambda snapshot, record, raw, fmt, rbargs:
-            _damo_fmt_str.format_nr(len(snapshot.regions), raw),
+            lambda snapshot, record, fmt:
+            _damo_fmt_str.format_nr(len(snapshot.regions), fmt.raw_number),
             'the number of regions in the snapshot'),
         Formatter('<region box colors>',
-            lambda snapshot, record, raw, fmt, rbargs:
-            _damo_ascii_color.color_samples(rbargs.colorset),
+            lambda snapshot, record, fmt:
+            _damo_ascii_color.color_samples(fmt.region_box_format.colorset),
             'available colors for the region box'),
         Formatter('<region box description>',
-            lambda snapshot, record, raw, fmt, rbargs:
-            rbargs.description_msg(raw),
+            lambda snapshot, record, fmt:
+            fmt.region_box_format.description_msg(fmt.raw_number),
             'description about region box (what and how it represents)'),
         Formatter('<temperature-sz histogram>',
-                  lambda snapshot, record, raw, fmt, rbargs:
-                  temperature_sz_hist_str(snapshot, record, raw, fmt),
+                  lambda snapshot, record, fmt:
+                  temperature_sz_hist_str(snapshot, record, fmt),
                   'temperature to total size of the regions histogram'),
         Formatter('<recency-sz histogram>',
-                  lambda snapshot, record, raw, fmt, rbargs:
-                  recency_hist_str(snapshot, record, raw, fmt),
+                  lambda snapshot, record, fmt:
+                  recency_hist_str(snapshot, record, fmt),
                   'last accessed time to total size of the regions histogram'),
         Formatter('<heatmap>',
-                  lambda snapshot, record, raw, fmt, rbargs:
-                  heatmap_str(snapshot, record, raw, fmt),
+                  lambda snapshot, record, fmt:
+                  heatmap_str(snapshot, record, fmt),
                   'heatmap of the snapshot'),
         ]
 
 region_formatters = [
         Formatter(
             '<index>',
-            lambda index, region, raw, fmt, rbargs:
-            _damo_fmt_str.format_nr(index, raw),
+            lambda index, region, fmt:
+            _damo_fmt_str.format_nr(index, fmt.raw_number),
             'index of the region in the regions of the snapshot'),
         Formatter(
             '<start address>',
-            lambda index, region, raw, fmt, rbargs:
-            _damo_fmt_str.format_sz(region.start, raw),
+            lambda index, region, fmt:
+            _damo_fmt_str.format_sz(region.start, fmt.raw_number),
             'start address of the region'),
         Formatter(
             '<end address>',
-            lambda index, region, raw, fmt, rbargs:
-            _damo_fmt_str.format_sz(region.end, raw),
+            lambda index, region, fmt:
+            _damo_fmt_str.format_sz(region.end, fmt.raw_number),
             'end address of the region'),
         Formatter(
             '<size>',
-            lambda index, region, raw, fmt, rbargs:
-            _damo_fmt_str.format_sz(region.size(), raw),
+            lambda index, region, fmt:
+            _damo_fmt_str.format_sz(region.size(), fmt.raw_number),
             'size of the region'),
         Formatter(
             '<access rate>',
-            lambda index, region, raw, fmt, rbargs:
-            _damo_fmt_str.format_percent(region.nr_accesses.percent, raw),
+            lambda index, region, fmt:
+            _damo_fmt_str.format_percent(region.nr_accesses.percent,
+                                         fmt.raw_number),
             'monitored access rate of the region'),
         Formatter(
             '<age>',
-            lambda index, region, raw, fmt, rbargs:
-            _damo_fmt_str.format_time_us(region.age.usec, raw),
+            lambda index, region, fmt:
+            _damo_fmt_str.format_time_us(region.age.usec, fmt.raw_number),
             'how long the access pattern of the region has maintained'),
         Formatter(
             '<temperature>',
-            lambda index, region, raw, fmt, rbargs:
-            temperature_str(region, raw, fmt),
+            lambda index, region, fmt:
+            temperature_str(region, fmt.raw_number, fmt),
             'access temperature of the region'),
         Formatter(
             '<box>',
-            lambda index, region, raw, fmt, rbargs:
-            rbargs.to_str(region),
+            lambda index, region, fmt:
+            fmt.region_box_format.to_str(region),
             'user-customizable (via --region_box_*) box (age/access_rate/size by default)'),
         ]
 
 default_record_head_format = 'kdamond <kdamond index> / context <context index> / scheme <scheme index> / target id <target id> / recorded for <duration> from <abs start time>'
-default_snapshot_head_format = 'monitored time: [<start time>, <end time>] (<duration>)'
+default_snapshot_head_format = 'monitored time: [<start time>, <end time>] (<duration>)\n<heatmap>'
+default_snapshot_head_format_without_heatmap = 'monitored time: [<start time>, <end time>] (<duration>)'
 default_region_format = '<index> addr <start address> size <size> access <access rate> age <age>'
 
-def temperature_sz_hist_str(snapshot, record, raw, fmt):
+def temperature_sz_hist_str(snapshot, record, fmt):
+    raw = fmt.raw_number
     if len(snapshot.regions) == 0:
         return 'no region in snapshot'
     hist = {}
@@ -205,7 +209,8 @@ def temperature_sz_hist_str(snapshot, record, raw, fmt):
         lines.append('%s %s %s' % (trange_str, sz_str, bar))
     return '\n'.join(lines)
 
-def recency_hist_str(snapshot, record, raw, fmt):
+def recency_hist_str(snapshot, record, fmt):
+    raw = fmt.raw_number
     if len(snapshot.regions) == 0:
         return 'no region in snapshot'
     hist = {}
@@ -280,6 +285,7 @@ def region_after(addr, regions):
 class HeatPixel:
     start = None
     end = None
+    total_heat = None
     temperature = None
     is_void = None
 
@@ -287,15 +293,16 @@ class HeatPixel:
         self.start = start
         self.end = end
 
-        self.temperature = 0
+        self.total_heat = 0
         self.is_void = True
         for region in regions:
             if region.end <= start:
                 continue
             if end <= region.start:
-                return
+                break
             self.is_void = False
             self.add_temperature(region, temperature_weights)
+        self.temperature = self.total_heat / (self.end - self.start)
 
     def add_temperature(self, region, weights):
         start = self.start
@@ -304,12 +311,12 @@ class HeatPixel:
         if start <= region.start and region.end <= end:
             # region is in the pixel
             pass
-        elif region.start < start and end < region.end:
+        elif region.start <= start and end <= region.end:
             # pixel is in the region
             region = copy.deepcopy(region)
             region.start = start
             region.end = end
-        elif region.end > start:
+        elif region.start < start:
             # region intersecting right part
             # <region>
             #    <pixel>
@@ -321,9 +328,10 @@ class HeatPixel:
             # <pixel>
             region = copy.deepcopy(region)
             region.end = end
-        self.temperature += temperature_of(region, weights)
+        self.total_heat += temperature_of(region, weights) * (region.size())
 
-def heatmap_str(snapshot, record, raw, fmt):
+def heatmap_str(snapshot, record, fmt):
+    raw = fmt.raw_number
     total_sz = 0
     for region in snapshot.regions:
         total_sz += region.size()
@@ -365,7 +373,12 @@ def heatmap_str(snapshot, record, raw, fmt):
                 (pixel.temperature - min_temperature) / temperature_unit)
         dots.append(_damo_ascii_color.colored(
             '%d' % temp_level, fmt.snapshot_heatmap_colorset, temp_level))
-    return ''.join(dots)
+    dots = ''.join(dots)
+    comment = '# min/max temperatures: %s, %s, column size: %s' % (
+            _damo_fmt_str.format_nr(min_temperature, raw),
+            _damo_fmt_str.format_nr(max_temperature, raw),
+            _damo_fmt_str.format_sz(sz_unit, raw))
+    return '%s\n%s' % (dots, comment)
 
 def rescale(val, orig_scale_minmax, new_scale_minmax, logscale=True):
     '''Return a value in new scale
@@ -491,7 +504,7 @@ class RegionBoxAttr:
         self.display_min_max = display_min_max
         self.display_logscale = display_logscale
 
-class RegionBox:
+class RegionBoxFormat:
     sorted_access_patterns = None
     length = None
     horizontal_align = None
@@ -608,31 +621,24 @@ def apply_min_chars(min_chars, field_name, txt):
             return txt
     return txt
 
-def format_template(template, formatters, min_chars, fmt, index, region,
-                    snapshot, record, raw, region_box_args):
+def format_output(template, formatters, fmt, record, snapshot=None,
+                  region=None, index=None):
     if template == '':
         return
     for formatter in formatters:
         if template.find(formatter.keyword) == -1:
             continue
         if formatters == record_formatters:
-            txt = formatter.format_fn(record, raw)
+            txt = formatter.format_fn(record, fmt.raw_number)
         elif formatters == snapshot_formatters:
-            txt = formatter.format_fn(snapshot, record, raw, fmt,
-                                      region_box_args)
+            txt = formatter.format_fn(snapshot, record, fmt)
         elif formatters == region_formatters:
-            txt = formatter.format_fn(index, region, raw, fmt, region_box_args)
-        txt = apply_min_chars(min_chars, formatter.keyword, txt)
+            txt = formatter.format_fn(index, region, fmt)
+        txt = apply_min_chars(fmt.min_chars_for, formatter.keyword, txt)
         template = template.replace(formatter.keyword, txt)
     template = template.replace('\\n', '\n')
     return template
 
-def format_output(template, formatters, min_chars, raw, region_box_args,
-                  fmt, record, snapshot=None, region=None, index=None):
-    if template == '':
-        return None
-    return format_template(template, formatters, min_chars, fmt, index, region,
-                           snapshot, record, raw, region_box_args)
 
 def temperature_of(region, weights):
     sz_weight, access_rate_weight, age_weight = weights
@@ -667,7 +673,7 @@ def sorted_regions(regions, sort_fields, sort_dsc_keys, temperature_weights):
 
 def fmt_records(fmt, records):
     sorted_access_patterns = SortedAccessPatterns(records)
-    region_box_args = RegionBox(sorted_access_patterns,
+    fmt.region_box_format = RegionBoxFormat(sorted_access_patterns,
             RegionBoxAttr(fmt.region_box_values[0],
                 fmt.region_box_min_max_length,
                 fmt.region_box_scales[0] == 'log'), fmt.region_box_align,
@@ -682,16 +688,13 @@ def fmt_records(fmt, records):
     for record in records:
         outputs.append(
                 format_output(
-                    fmt.format_record_head, record_formatters,
-                    fmt.min_chars_for, fmt.raw_number, region_box_args, fmt,
-                    record))
+                    fmt.format_record_head, record_formatters, fmt, record))
         snapshots = record.snapshots
 
         for sidx, snapshot in enumerate(snapshots):
             outputs.append(
                     format_output(
                         fmt.format_snapshot_head, snapshot_formatters,
-                        fmt.min_chars_for, fmt.raw_number, region_box_args,
                         fmt, record, snapshot))
             for r in snapshot.regions:
                 r.nr_accesses.add_unset_unit(record.intervals)
@@ -702,21 +705,17 @@ def fmt_records(fmt, records):
                 outputs.append(
                         format_output(
                             fmt.format_region, region_formatters,
-                            fmt.min_chars_for, fmt.raw_number,
-                            region_box_args, fmt, record, snapshot, r, idx))
+                            fmt, record, snapshot, r, idx))
             outputs.append(
                     format_output(
                         fmt.format_snapshot_tail, snapshot_formatters,
-                        fmt.min_chars_for, fmt.raw_number, region_box_args,
                         fmt, record, snapshot))
 
             if sidx < len(snapshots) - 1 and not fmt.total_sz_only():
                 outputs.append('')
         outputs.append(
                 format_output(
-                    fmt.format_record_tail, record_formatters,
-                    fmt.min_chars_for, fmt.raw_number, region_box_args, fmt,
-                    record))
+                    fmt.format_record_tail, record_formatters, fmt, record))
     outputs = [o for o in outputs if o is not None]
     return '\n'.join(outputs)
 
@@ -788,6 +787,9 @@ class RecordsVisualizationFormat:
     region_box_scales = None
     region_box_align = None
 
+    # RegionBoxFormat.  Set on fmt_records()
+    region_box_format = None
+
     min_chars_for = None
     raw_number = None
     json = None
@@ -903,6 +905,7 @@ class RecordsVisualizationFormat:
 
 def set_formats(args):
     if args.style == 'simple-boxes':
+        args.format_snapshot_head = default_snapshot_head_format_without_heatmap
         args.format_region = '<box> size <size> access rate <access rate> age <age>'
         args.region_box_min_max_height = [1, 1]
         args.region_box_min_max_length = [1, 40]
